package protocolP2P;
import exception.ProtocolError;
import exception.VersionError;
import exception.SizeError;
import exception.InternalError;
import exception.TransmissionError;
import protocolP2P.Payload;
import protocolP2P.RequestResponseCode;
import java.util.ArrayList;
import java.lang.Byte;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.io.IOException;
import java.net.UnknownHostException;

/** Representation of datagram.
 * @author	Louis Royer
 * @author	Flavien Haas
 * @author	JS Auge
 * @version	1.0
 */
public class ProtocolP2PDatagram {
	private final static byte PROTOCOL_VERSION = 0x11;
	private final static int VERSION_POSITON = 0;
	private byte version;
	private Payload payload;

	/** Constructor with payload parameter (typically used when sending datagram).
	 * @param payload	the payload associated with the datagram to send
	 */
	public ProtocolP2PDatagram(Payload payload) {
		version = PROTOCOL_VERSION;
		this.payload = payload;
	}

	/** Send datagram on socket (from client)
	 * @param socket	DatagramSocket used to send datagram.
	 * @param host	host to send datagram (null if this is a response)
	 * @param port	port to send datagram (null if this is a response)
	 * @throws	InternalError
	 * @throws	UnknownHostException
	 * @throws	IOException
	 */
	public void send(DatagramSocket socket, String host, int port) throws InternalError, UnknownHostException, IOException {
		InetAddress dst = InetAddress.getByName(host);
		byte[] datagram = toDatagram();
		// generate DatagramPacket
		DatagramPacket datagramPacket = new DatagramPacket(datagram, datagram.length, dst, port);
		// send it
		socket.send(datagramPacket);
	}

	/** Send datagram on socket (from server, as a response)
	 * @param socket	DatagramSocket used to send datagram.
	 * @throws	InternalError
	 * @throws	IOException
	 */
	public void send(DatagramSocket socket) throws InternalError, IOException {
		byte[] datagram = toDatagram();
		// generate DatagramPacket
		DatagramPacket datagramPacket = new DatagramPacket(datagram, datagram.length);
		socket.send(datagramPacket);
	}

	/** Receive datagram on socket
	 * @param socket	DatagramSocket used to receive datagram
	 * @return	payload of the datagram
	 * @throws	TransmissionError
	 * @throws	ProtocolError
	 * @throws	VersionError
	 * @throws	InternalError
	 * @throws	SizeError
	 */
	public static Payload receive(DatagramSocket socket) throws TransmissionError, ProtocolError, VersionError, InternalError, SizeError {
		// reception
		byte[] datagram = new byte[4096];
		DatagramPacket reception = new DatagramPacket(datagram, datagram.length);
		// contruction
		ProtocolP2PDatagram p = new ProtocolP2PDatagram(datagram);
		return p.getPayload();
	}
	/** Private constructor with datagram as byte[] parameter (typically used when receiving datagram).
	 * @param datagram	the full datagram received
	 * @throws	TransmissionError
	 * @throws	ProtocolError
	 * @throws	VersionError
	 * @throws	InternalError
	 * @throws	SizeError
	 */
	private ProtocolP2PDatagram(byte[] datagram) throws TransmissionError, ProtocolError, VersionError, InternalError, SizeError {
		// unwrap version
		version = datagram[VERSION_POSITON];
		checkProtocolVersion(); // this can throw VersionError
		RequestResponseCode r = RequestResponseCode.fromCode(datagram[RequestResponseCode.RRCODE_POSITION]); // this can throw ProtocolError
		switch (r) {
			case LIST_RESPONSE:
				payload = (Payload) new FileList(datagram);
				break;
			case LOAD_RESPONSE:
				payload = (Payload) new FilePart(datagram);
				break;
			default:
				payload = new Payload(datagram); // this can throw TransmissionError
				break;
			}
	}

	/** Returns a byte[] containing full datagram (typically used when sending datagram).
	 * This datagram is still complete and ready to be send.
	 * @return the full datagram to send
	 * @throws	InternalError
	 */
	private byte[] toDatagram() throws InternalError {
		byte[] datagram = payload.toDatagram();
		datagram[VERSION_POSITON] = version;
		return datagram;
		
	}

	/** Returns Payload associated with the datagram.
	 * @return	payload associated with the datagram
	 */
	private Payload getPayload() {
		return payload;
	}

	/** Used to check protocol version when a datagram is constructed from bytes[].
	 * @throws	VersionError
	 */
	private void checkProtocolVersion() throws VersionError {
		if (PROTOCOL_VERSION != version) {
			throw new VersionError();
		}
	}
}

